// Simple program for testing the capabilities of soundcards, based on the ALSA API.
//
// Code by
//   Sabine Cremer <sc@dl1dbc.net>
//
// Compile with option "-lasound".
//
// Downloaded from http://www.george-smart.co.uk/
//   Hosted with permission.

#include <stdio.h>
#include <stdlib.h>
#include <alsa/asoundlib.h>

static int p_avail = 1;
static int c_avail = 1;


void hw_capabilities (snd_pcm_t *handle,
		      snd_pcm_hw_params_t *hw_params,
		      char *device)
{
	snd_pcm_format_t *format;
	char *match = "plug";
	unsigned int channels = 10;
	unsigned int samplerate[] = {8000, 16000, 24000, 44100, 48000, 96000, 192000};
	int testcases = sizeof(samplerate)/sizeof(unsigned int);
	unsigned int lastrate;
	
	int err;
	int i, k;

/* parameter evaluation seems to fail if plughw used */
	if (strstr(device, match)) { 
		printf("No evaluation possible for plughw devices; try hw instead!");
	} else {

/* initialize configuration structure with full configuration space */
		if ((err = snd_pcm_hw_params_any (handle, hw_params)) < 0) {
			fprintf (stderr, "cannot initialize audio device configuration structure (%s)\n",
				 snd_strerror (err));
			exit (1);
		}

		printf("Available sample rates:\n");

		for (i = 0; i < testcases; i++) {

/* get available sample rates of the device (use snd_pcm_hw_params_any first **any time** before!) */
			if ((err = snd_pcm_hw_params_any (handle, hw_params)) < 0) {
				fprintf (stderr, "cannot initialize audio device configuration structure (%s)\n",
					 snd_strerror (err));
				exit (1);
			} 

			if ((err = snd_pcm_hw_params_set_rate_near (handle, hw_params, &samplerate[i], 0)) < 0) {
				fprintf (stdout, "cannot set sample rate (%s)\n",
					 snd_strerror (err));
				exit (1);
			}

			if (samplerate[i] != lastrate) {
				printf("\t%i Hz\n", samplerate[i]);

/* is needed to evaluate maximum channel count for each sample rate: */
				if ((err = snd_pcm_hw_params_set_channels_last(handle, hw_params, &channels)) < 0) {
					fprintf (stdout, "cannot restrict channels to maximum (%s)\n",
						 snd_strerror (err));
					exit (1);
				} 

/* is needed to get the highest available resolution: */
		if ((err = snd_pcm_hw_params_set_format_last(handle, hw_params, format)) < 0) {
			fprintf (stdout, "cannot restrict format to last (%s)\n",
				 snd_strerror (err));
			exit (1);
		} 

/* write the prepared (snd_pcm_hw_params_set_...) hardware configuration to the device */
/* never set further parameters and call snd_pcm_hw_parameters again - it does not work! */
				if ((err = snd_pcm_hw_params (handle, hw_params)) < 0) {
					fprintf (stderr, "(snd_pcm_hw_params): cannot set parameters (%s)\n",
						 snd_strerror (err));
					exit (1);
				}

/* get maximum channel count for sample rate (use snd_pcm_hw_params_any, **snd_pcm_hw_params_set_channels_last** first!) */
				if ((err = snd_pcm_hw_params_get_channels_max (hw_params, &channels)) < 0) {
					fprintf (stdout, "cannot get maximum channel count (%s)\n",
						 snd_strerror (err));
					exit (1);
				}
				printf("\t\t%i channel", channels);
				switch(channels) {
					case 1: printf(" (mono)"); break;
					case 2: printf("s (stereo)"); break;
					case 6: printf("s (5.1)"); break;
					case 8: printf("s (7.1)"); break;
					case 10: printf("s (8 + 2)");
				}
				printf("\n");

/* get sample resolution info (use snd_pcm_hw_params first) */
				if ((err = snd_pcm_hw_params_get_sbits (hw_params)) < 0) {
					fprintf (stderr, "sbits  (%s)\n",
						 snd_strerror (err));
					exit (1);
				}
				printf("\t\t%i bit max. resolution\n", err);

/* get supported sample formats */
					printf("\t\t\tSupported sample formats: ");
					for (k= 0; k < SND_PCM_FORMAT_LAST; ++k) {
						err = snd_pcm_hw_params_test_format(handle, hw_params, k);
						const char *s = snd_pcm_format_name(k); 
						const char *d = snd_pcm_format_description(k);
						if (err == 0) 
							printf("\n\t\t\t\t%s (%s)  ", s, d); 
					} 
					printf("\n");
				
/* get supported access types */
				printf("\t\t\tSupported access types: ");
				for (k= 0; k < SND_PCM_ACCESS_LAST; ++k) {
					err = snd_pcm_hw_params_test_access(handle, hw_params, k);
					const char *s = snd_pcm_access_name(k); 
					if (err == 0)
						printf("\n\t\t\t\t%s", s); 
				}
				printf("\n");

			} /* if (samplerate[i] != lastrate */

			lastrate = samplerate[i];
		} /* for (i = 0; i < testcases; i++) */


		printf("\n");


	} /* if (!strstr(device, match)) */

}


/***********************************************************************************************/
int main (int argc, char *argv[])
/***********************************************************************************************/
{
	char *pdevice = "hw:0,0";
	char *cdevice = "hw:0,0";
	int err;
	snd_pcm_t *phandle;
	snd_pcm_t *chandle;
	snd_pcm_hw_params_t *phw_params;
	snd_pcm_hw_params_t *chw_params;
/*	snd_pcm_status_t *status; */
	snd_output_t *output = NULL;

	if (argc > 1) {
		pdevice = argv[1];
		cdevice = argv[1];
	}

/* get PCM handle (playback) */
	printf ("\nTrying to open audio device %s for playback (snd_pcm_open)\n", pdevice);
	if ((err = snd_pcm_open (&phandle, pdevice, SND_PCM_STREAM_PLAYBACK, 0)) < 0) 
		p_avail = 0;

/* get PCM handle (capture) */
	printf ("Trying to open audio device %s for capture (snd_pcm_open)\n", cdevice);
	if ((err = snd_pcm_open (&chandle, cdevice, SND_PCM_STREAM_CAPTURE, 0)) < 0) 
		c_avail = 0;

/* get handle for dump output */
	err = snd_output_stdio_attach(&output, stdout, 0);

/* dump hardware information */
	snd_pcm_dump(chandle, output);

/* dump status information 
	printf("PCM status:\n");
	if ((err = snd_pcm_status (chandle, status)) < 0) {
		fprintf (stderr, "stream status error (%s)\n", 
			 snd_strerror (err));
		exit (1);
	}
	snd_pcm_status_dump(status, output);

*/

/* validate memory allocation (playback) */
	if ((err = snd_pcm_hw_params_malloc (&phw_params)) < 0) {
		fprintf (stderr, "cannot allocate hardware playback configuration structure (%s)\n",
			snd_strerror (err));
	}

/* validate memory allocation (capture) */
	if ((err = snd_pcm_hw_params_malloc (&chw_params)) < 0) {
		fprintf (stderr, "cannot allocate hardware capture configuration structure (%s)\n",
			 snd_strerror (err));
		exit (1);
	}


/************************************************************************************************/
/* evaluation of hardware capabilities (playback) */
	printf("\n*** Playback capabilities\n");
	if (!p_avail) {
		printf("%s cannot be opened for playback\n", pdevice);
	} else {
		hw_capabilities(phandle, phw_params, pdevice); 
	}

/* evaluation of hardware capabilities (capture) */
	printf("\n*** Capture capabilities\n");
	if (!c_avail) {
		printf("%s cannot be opened for capture\n", cdevice);
	} else {
		hw_capabilities(chandle, chw_params, cdevice); 
	}
/************************************************************************************************/

	if (c_avail) {
/*  snd_pcm_prepare is called automatically by snd_pcm_hw_params (!) */
		printf("\nPreparing device %s for operation (snd_pcm_prepare): ", cdevice);
		if ((err = snd_pcm_prepare (chandle)) < 0) {
			fprintf (stderr, "cannot prepare audio interface for use (%s)\n",
				 snd_strerror (err));
			exit (1);
		}
		printf("ok\n");

/*  start selected operation (playback or capture) */
		printf("Starting PCM operation (snd_pcm_start): ");
		if ((err = snd_pcm_start (chandle)) < 0) {
			fprintf (stderr, "cannot start operation of audio interface (%s)\n",
				 snd_strerror (err));
			exit (1);
		}
		printf("ok\n");

/*  stop PCM operation */
		printf("Stopping PCM operation (snd_pcm_drop): ");
		if ((err = snd_pcm_drop (chandle)) < 0) {
			fprintf (stderr, "cannot stop operation of audio interface (%s)\n",
				 snd_strerror (err));
			exit (1);
		}
		printf("ok\n\n");
	}

	snd_pcm_hw_params_free (chw_params); 
	snd_pcm_hw_params_free (phw_params); 
	if (c_avail) snd_pcm_close (chandle); 
	if (p_avail) snd_pcm_close (phandle); 
	exit (0);
}
